package termopl;

import java.io.*;
import java.util.*;

import pl.sgjp.morfeusz.*;

public class Tagger 
{
	
	private Morfeusz morfeusz;
	private File input, output;
	
	public Tagger(File input, File output)
	{
		morfeusz = Morph.getAnalyzer();
		morfeusz.setWhitespaceHandling(WhitespaceHandling.APPEND_WHITESPACES);
		this.input = input;
		this.output = output;
	}

	public void run()
	{
		String cmd;
		String err = null;
		
		File dir = new File(TermoPL.appLocation + File.separator + "concraft-pl");
		
		if (dir.exists()) {
			String path = dir.getAbsolutePath() + File.separator + "concraft-pl";
			
			if (TermoPL.isWindows) path += ".exe";
			
			File file = new File(path);
			
			if (file.exists()) {
				File[] models = dir.listFiles(new FilenameFilter() {
				    public boolean accept(File dir, String name) 
				    {
				        return name.startsWith("concraft-pl-model") && name.endsWith(".gz");
				    }
				});
				
				if (models.length > 0) {
					cmd = file.getAbsolutePath() + " tag " + models[0].getAbsolutePath();
				}
				else {
					cmd = null;
					err = "Can't find a model for tagging.";
				}
			}
			else {
				cmd = null;
				err = "Can't find a tagger";
			}
		}
		else {
			cmd = null;
			err = "Can't find a tagger";
		}
		
		if (err == null) {
			try {
				File tmp1, tmp2;
				
				tmp1 = File.createTempFile("termopl", null);
				tmp2 = File.createTempFile("termopl", null);
				tmp1.deleteOnExit();
				tmp2.deleteOnExit();
				cmd += " -i " + tmp1.getAbsolutePath() + " -o " + tmp2.getAbsolutePath();
				prepareInput(input, tmp1);
				tag(cmd);
				prepareOutput(tmp2, output);
				tmp1.delete();
				tmp2.delete();
			}
			catch (Exception exception) {
				exception.printStackTrace();
			}
		}
		else {
			System.out.println(err);
		}
	}
	
	public void prepareInput(File in, File out)
	{
		BufferedReader textReader;
		PrintWriter pw;
		
		try {
			textReader = new BufferedReader(new InputStreamReader(new FileInputStream(in), "UTF8"));
			pw = new PrintWriter(out, "UTF8");
			StringBuffer buffer = new StringBuffer();
			String line;

			line = textReader.readLine();
			while (line != null) {
				if (line.trim().isEmpty()) {
					saveParagraph(buffer, pw);
				}
				else {
					if (buffer.length() > 0) buffer.append(' ');
					buffer.append(line);
				}
				line = textReader.readLine();
			}
			saveParagraph(buffer, pw);
			pw.flush();
			pw.close();
			textReader.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		} 
	}
	
	public void saveParagraph(StringBuffer buffer, PrintWriter pw)
	{
		if (buffer.length() > 0) {
			String par = buffer.toString();
			List<MorphInterpretation> interps = morfeusz.analyseAsList(par);
			LinkedList<Integer> npsNodes = new LinkedList<Integer>();
			boolean nps;
			
			for (MorphInterpretation mi : interps) {
				String orth = mi.getOrth();
				int startNode = mi.getStartNode();
				
				if (npsNodes.contains(startNode)) nps = true;
				else nps = false;
				if (orth.equals(orth.trim())) {
					int endNode = mi.getEndNode();
					
					if (!npsNodes.contains(endNode)) npsNodes.addFirst(endNode);
				}
				saveInterp(mi, nps, pw);
			}
			pw.print('\n');
			buffer.setLength(0);
		}
	}
	
	public void saveInterp(MorphInterpretation interp, boolean nps, PrintWriter pw)
	{
		int startNode = interp.getStartNode();
		int endNode = interp.getEndNode();
		String orth = interp.getOrth().trim();
		String lemma = interp.getLemma();
		String tag = interp.getTag(morfeusz);
		
		if (!lemma.equals(":")) lemma = lemma.split(":")[0];
		if (tag.contains(".")) {
			LinkedList<String> tags = splitTag(tag);
			
			for (String t : tags) {
				saveInterp(pw, startNode, endNode, orth, lemma, t, nps);
			}
		}
		else saveInterp(pw, startNode, endNode, orth, lemma, tag, nps);
	}
	
	public void saveInterp(PrintWriter pw, int startNode, int endNode, String orth, String lemma, String tag, boolean nps)
	{
		pw.print(startNode);
		pw.print('\t');
		pw.print(endNode);
		pw.print('\t');
		pw.print(orth);
		pw.print('\t');
		pw.print(lemma);
		pw.print('\t');
		pw.print(tag);
		pw.print('\t');
		pw.print('\t');
		pw.print('\t');
		pw.print("0.000");
		pw.print('\t');
		pw.print('\t');
		pw.print('\t');
		if (nps) pw.print("nps");
		pw.print('\n');
	}
	
	public LinkedList<String> splitTag(String tag)
	{
		LinkedList<String> tags = new LinkedList<String>();
		String[] cats = tag.split(":");
		
		splitTag(tags, "", cats, 0);
		return tags;
	}
	
	public void splitTag(LinkedList<String> tags, String tag, String[] cats, int index)
	{
		if (index < cats.length) {
			if (cats[index].contains(".")) {
				String[] markers = cats[index].split("\\.");
				
				for (String m : markers) {
					if (tag.isEmpty()) splitTag(tags, m, cats, index + 1);
					else splitTag(tags, tag + ":" + m, cats, index + 1);
				}
			}
			else {
				if (tag.isEmpty()) splitTag(tags, cats[index], cats, index + 1);
				else splitTag(tags, tag + ":" + cats[index], cats, index + 1);
			}
		}
		else tags.add(tag);
	}
	
	public void prepareOutput(File in, File out)
	{
		BufferedReader textReader;
		PrintWriter pw;
		
		try {
			textReader = new BufferedReader(new InputStreamReader(new FileInputStream(in), "UTF8"));
			pw = new PrintWriter(out, "UTF8");
			String line;
			String startNode = "", endNode = "";

			line = textReader.readLine();
			while (line != null) {
				if (!line.trim().isEmpty()) {
					String[] cols = line.split("\t");

					if (cols.length == 12 && cols[11].equals("disamb")) {
						if (!cols[0].equals(startNode) || !cols[1].equals(endNode)) {
							startNode = cols[0];
							endNode= cols[1];
							pw.print(cols[2]);
							pw.print('\t');
							pw.print(cols[3]);
							pw.print('\t');
							pw.print(cols[4]);
							if (cols[10].equals("nps")) {
								pw.print('\t');
								pw.print("nps");
							}
							pw.println();
							if (cols[9].equals("eos")) pw.println("eos");
						}
					}
				}
				line = textReader.readLine();
			}
			pw.flush();
			pw.close();
			textReader.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		} 
		
	}
	
	public void tag(String cmd)
	{
		Runtime r = Runtime.getRuntime();
		
		try {
			Process p = r.exec(cmd);
			p.waitFor();
		} 
		catch (Exception e) {
			e.printStackTrace();
		}
	}

}
